from pathlib import Path

import tomllib

workspace = Path(__file__).parent.parent.parent.parent.parent

with (workspace / "Cargo.toml").open("rb") as f:
    cargo = tomllib.load(f)
    package = cargo["workspace"]["package"]
    author = ", ".join(
        author.replace("<", "(").replace(">", ")") for author in package["authors"]
    )

with (workspace / "crates/pyndakaas/Cargo.toml").open("rb") as f:
    cargo = tomllib.load(f)
    package = cargo["package"]
    release = package["version"]
    description = package["description"]
    project = package["name"]
    copyright = f"2024-2025, {author}"

version = release

extensions = [
    "sphinx.ext.autodoc",
    "sphinx_rtd_theme",
]

autodoc_typehints = "description"  # show type hints from signature in docs

templates_path = ["_templates"]
exclude_patterns = []

html_theme = "sphinx_rtd_theme"
# html_static_path = ['_static']

html_logo = "../../../../assets/logo.svg"
pygments_style = "sphinx"

# html_theme_options = {
#     'logo_only': True,
# }



import pindakaas.solver

from .encoding import CNF, WCNF, ClauseDatabase, Constraint
from .pindakaas import Encoder, Formula, InvalidEncoder, Lit, Unsatisfiable

__doc__ = pindakaas.__doc__
__all__ = [
    "Constraint",
    "ClauseDatabase",
    "CNF",
    "Encoder",
    "Formula",
    "Lit",
    "WCNF",
    "Unsatisfiable",
    "InvalidEncoder",
]



import pindakaas


def test_bool_lin():
    f = pindakaas.CNF()
    x, y, z = f.new_vars(3)
    assert str(x + 2) == "x₁ + 2", "__add__ incorrect"
    assert str(2 + x) == "x₁ + 2", "__radd__ incorrect"
    assert str(x - 2) == "x₁ + -2", "__sub__ incorrect"
    assert str(2 * x) == "2*x₁", "__mul__ incorrect"
    assert str(x * 2) == "2*x₁", "__rmul__ incorrect"
    assert str(x + y - z + 2) == "-x₃ + x₂ + x₁ + 2"
    c = sum([x, y, z])
    assert str(c) == "x₃ + x₂ + x₁"
    c *= 2
    assert str(c) == "2*x₃ + 2*x₂ + 2*x₁"
    c = x + y + z
    d = c == 2
    assert str(d) == "x₃ + x₂ + x₁ == 2"
    d = c < 2
    assert str(d) == "x₃ + x₂ + x₁ <= 1"
    d = c >= 2
    assert str(d) == "x₃ + x₂ + x₁ >= 2"


def test_bool_lin_ops():
    f = pindakaas.CNF()
    x, y, z = f.new_vars(3)
    f = x + y
    g = x - z
    assert str(f + 2) == "x₂ + x₁ + 2", "__add__ incorrect"
    assert str(2 + f) == "x₂ + x₁ + 2", "__radd__ incorrect"
    assert str(f + g) == "-x₃ + x₁ + x₂ + x₁", "__add__ incorrect"
    assert str(f * 2) == "2*x₂ + 2*x₁", "__mul__ incorrect"
    assert str(2 * f) == "2*x₂ + 2*x₁", "__rmul__ incorrect"


def test_formula_ops():
    f = pindakaas.CNF()
    x, y, z = f.new_vars(3)
    f = x & y
    g = x | z
    assert str(f & True) == "x₁ ∧ x₂ ∧ true", "__and__ incorrect"
    assert str(True & f) == "x₁ ∧ x₂ ∧ true", "__rand__ incorrect"
    assert str(f & g) == "x₁ ∧ x₂ ∧ (x₁ ∨ x₃)", "__and__ incorrect"
    assert str(f | True) == "(x₁ ∧ x₂) ∨ true", "__or__ incorrect"
    assert str(True | f) == "(x₁ ∧ x₂) ∨ true", "__ror__ incorrect"
    assert str(f | g) == "x₁ ∨ x₃ ∨ (x₁ ∧ x₂)", "__or__ incorrect"
    assert str(f ^ True) == "(x₁ ∧ x₂) ⊻ true", "__xor__ incorrect"
    assert str(True ^ f) == "(x₁ ∧ x₂) ⊻ true", "__rxor__ incorrect"
    assert str(f ^ g) == "(x₁ ∧ x₂) ⊻ (x₁ ∨ x₃)", "__xor__ incorrect"


def test_lit_ops():
    f = pindakaas.CNF()
    x, y = f.new_vars(2)
    assert str(x & True) == "x₁ ∧ true", "__and__ incorrect"
    assert str(True & x) == "x₁ ∧ true", "__rand__ incorrect"
    assert str(x | True) == "x₁ ∨ true", "__or__ incorrect"
    assert str(True | x) == "x₁ ∨ true", "__ror__ incorrect"
    assert str(x ^ True) == "x₁ ⊻ true", "__xor__ incorrect"
    assert str(True ^ x) == "x₁ ⊻ true", "__rxor__ incorrect"
    assert str(~x | ~y) == "¬x₁ ∨ ¬x₂"
    assert str(x & y) == "x₁ ∧ x₂"
    assert str(x ^ y) == "x₁ ⊻ x₂"
    assert str(x == y) == "x₁ ≡ x₂"



"""A module containing solvers and solving related classes."""

from abc import ABC, abstractmethod
from contextlib import contextmanager
from datetime import timedelta
from typing import ContextManager, Iterable, Iterator, Optional

from .encoding import ClauseDatabase, Constraint
from .pindakaas import Encoder, Lit
from .pindakaas.solver import CaDiCaLInner, KissatInner, Status


class Result(ABC):
    """The Result object returned after calling `solve()`.

    It gives access to e.g. solver status and the values of variables.
    """

    @property
    @abstractmethod
    def status(self) -> Status:
        """Return result from solving the database."""
        ...

    @abstractmethod
    def value(self, lit: Lit) -> Optional[bool]:
        """Return value for literal `lit`, or `None` if `lit` is assigned.

        :param lit: the literal for which to return the value

        :return: the value of `lit` if assigned

        """
        ...

    @abstractmethod
    def failed(self, lit: Lit) -> Optional[bool]:
        """Check if the given assumption literal was used to prove the unsatisfiability.

        The unsatisfiability of the formula is under the assumptions used for the last
        SAT search. Note also that for literals `lit` which are not assumption literals,
        the behavior of is not specified.

        :param lit: the assumption literal for which to return whether it contributed to
            the unsatisfiable result

        :return: whether `lit` contributed to the unsatisfiable result
        """
        ...


class Solver(ClauseDatabase):
    """An abstract class which extends a `ClauseDatabase` with solving capabilities."""

    def _set_time_limit(self, limit: Optional[timedelta]):
        if limit is not None:
            raise NotImplementedError("Solver does not support setting a time limit")

    @abstractmethod
    def _solve_assuming(self, assumptions: Iterable[Lit]) -> ContextManager[Result]: ...

    @contextmanager
    def solve(
        self,
        assumptions: Optional[Iterable[Lit]] = None,
        time_limit: Optional[timedelta] = None,
    ) -> Iterator[Result]:
        """Solve the current `ClauseDatabase`.

        :param assumptions: an optional iterable of assumptions literals which must hold
            for this solve call
        :param time_limit: an optional time limit before which the solver is terminated
            and the result is Unknown
        """
        self._set_time_limit(time_limit)
        assumptions = assumptions if assumptions is not None else []
        try:
            with self._solve_assuming(assumptions) as result:
                yield result
        finally:
            self._set_time_limit(None)


class CaDiCaL(Solver):
    """The `CaDiCaL <https://github.com/arminbiere/cadical>`_ SAT solver."""

    _inner: CaDiCaLInner

    def __init__(self):
        """Initialize solver."""
        self._inner = CaDiCaLInner()

    def _set_time_limit(self, limit: Optional[timedelta]):
        return self._inner.set_time_limit(limit)

    @contextmanager
    def _solve_assuming(self, assumptions: Iterable[Lit]) -> Iterator[Result]:
        # TODO: Investigate whether it is possible to avoid copying the solution
        (status, mapping) = self._inner.solve_assuming(assumptions)
        yield MapResult(status, mapping)

    def add_clause(self, clause: Iterable[Lit]):
        return self._inner.add_clause(iter(clause))

    def add_encoding(
        self,
        constraint: Constraint,
        encoder: Optional[Encoder] = None,
        conditions: Optional[Iterable[Lit]] = None,
    ):
        conditions = list(conditions) if conditions is not None else []
        return self._inner.add_encoding(constraint, encoder, conditions)

    def new_var_range(self, n: int):
        return self._inner.new_var_range(n)


class Kissat(Solver):
    """The `Kissat <https://github.com/arminbiere/kissat>`_ SAT solver."""

    _inner: KissatInner

    def __init__(self):
        self._inner = KissatInner()

    def _set_time_limit(self, limit: Optional[timedelta]):
        return self._inner.set_time_limit(limit)

    @contextmanager
    def _solve_assuming(self, assumptions: Iterable[Lit]) -> Iterator[Result]:
        # TODO: Investigate whether it is possible to avoid copying the solution
        (status, mapping) = self._inner.solve_assuming(assumptions)
        yield MapResult(status, mapping)

    def add_clause(self, clause: Iterable[Lit]):
        return self._inner.add_clause(iter(clause))

    def add_encoding(
        self,
        constraint: Constraint,
        encoder: Optional[Encoder] = None,
        conditions: Optional[Iterable[Lit]] = None,
    ):
        conditions = list(conditions) if conditions is not None else []
        return self._inner.add_encoding(constraint, encoder, conditions)

    def new_var_range(self, n: int):
        return self._inner.new_var_range(n)


class MapResult(Result):
    def __init__(self, status: Status, mapping: dict[int, bool]):
        self._status = status
        self._mapping = mapping

    @property
    def status(self) -> Status:
        return self._status

    def value(self, lit: Lit) -> Optional[bool]:
        if self.status == Status.SATISFIED:
            return self._mapping.get(int(lit))
        return None

    def failed(self, lit: Lit) -> Optional[bool]:
        if self.status == Status.UNSATISFIABLE:
            return self._mapping.get(int(lit))
        return None



#!/usr/bin/env python
import pindakaas

# Instantiate solver, create two new literals, and encode a XOR constraint
slv = pindakaas.solver.CaDiCaL()
x, y, z = slv.new_vars(3)
slv += x ^ y

with slv.solve(assumptions=[x]) as result:
    assert result.value(x) is True
    assert result.value(y) is False

with slv.solve(assumptions=[x, y]) as result:
    assert result.status == pindakaas.solver.Status.UNSATISFIABLE



from typing import Iterable

import pytest
from pindakaas import (
    CNF,
    WCNF,
    ClauseDatabase,
    Encoder,
    InvalidEncoder,
    Lit,
    Unsatisfiable,
)
from pindakaas.encoding import VarRange


class CustomDB(ClauseDatabase):
    clauses: list[list[int]]
    next_var: int

    def __init__(self):
        self.clauses = []
        self.next_var = 1

    def add_clause(self, clause: Iterable[Lit]):
        self.clauses.append([int(lit) for lit in clause])

    def new_var_range(self, n: int) -> VarRange:
        start = self.next_var
        self.next_var += n
        return VarRange(Lit.from_raw(start), Lit.from_raw(self.next_var - 1))


def test_unsat():
    f = CNF()
    with pytest.raises(Unsatisfiable):
        f.add_clause([])


def test_cnf():
    f = CNF()
    x, y = f.new_vars(2)
    assert list(f.variables()) == [x, y]
    f.add_clause([x, y])
    assert f.to_dimacs() == "p cnf 2 1\n1 2 0\n"
    assert f.clauses() == [[x, y]]


def test_encode_bool_lin_unsat():
    f = CNF()
    x, y, z = f.new_vars(3)
    with pytest.raises(Unsatisfiable):
        f += x * 3 + y * 2 + z >= 10


def test_invalid_encoder():
    f = CNF()
    x, y, z = f.new_vars(3)
    with pytest.raises(InvalidEncoder):
        f.add_encoding(x * 3 + y * 2 + z >= 3, encoder=Encoder.PAIRWISE)


def test_encode_bool_lin_default():
    f = CNF()
    x, y, z = f.new_vars(3)
    f += x * 3 + y * 2 + z >= 3
    x, y, z = f.new_vars(3)
    f.add_encoding(x + y + z == 1)
    assert (
        f.to_dimacs()
        == """p cnf 11 13
2 5 0
3 4 0
3 2 6 0
-4 1 0
-5 1 0
-6 1 0
7 8 9 0
-7 -10 0
-7 -11 0
-8 10 0
-8 -11 0
-9 -10 0
-9 11 0
"""
    )


def test_encode_formula():
    f = CNF()
    x, y, z = f.new_vars(3)
    f += x ^ z
    f.add_encoding(x == y, Encoder.TSEITIN)
    f.add_encoding(x & y)
    assert f.to_dimacs() == "p cnf 3 6\n1 3 0\n-1 -3 0\n-1 2 0\n1 -2 0\n1 0\n2 0\n"


def test_wcnf():
    f = WCNF()
    x, y = f.new_vars(2)
    f.add_clause([x, y])
    f.add_weighted_clause([x], 1)
    f.add_weighted_clause([y], 2)
    assert f.to_dimacs() == "p wcnf 2 3 4\n4 1 2 0\n1 1 0\n2 2 0\n"


def test_conditions():
    f = CNF()
    x, y, p = f.new_vars(3)
    f.add_encoding(x ^ y, conditions=[p])
    assert f.to_dimacs() == "p cnf 3 2\n3 1 2 0\n3 -1 -2 0\n"


def test_custom_db():
    f = CustomDB()
    assert f.new_var() == Lit.from_raw(1)
    x, y, p = f.new_vars(3)
    assert [x, y, p] == [
        Lit.from_raw(2),
        Lit.from_raw(3),
        Lit.from_raw(4),
    ]

    f.add_encoding(x ^ y, conditions=[p])
    assert f.clauses == [
        [4, 2, 3],
        [4, -2, -3],
    ]



#!/usr/bin/env python
import pindakaas

# Instantiate solver, create two new literals, and encode a XOR constraint
slv = pindakaas.solver.CaDiCaL()
x, y, z = slv.new_vars(3)
slv += x ^ y

# Solve and inspect status and literal values
with slv.solve() as result:
    assert result.status == pindakaas.solver.Status.SATISFIED
    assert result.value(x) != result.value(y)



from abc import ABC, abstractmethod
from typing import Iterable, Optional

from .pindakaas import (
    CNFInner,
    Encoder,
    Formula,
    Lit,
    VarRange,
    WCNFInner,
    _wrap_encode_constraint,
)

Constraint = Formula


class ClauseDatabase(ABC):
    """The abstract class to represent objects to which we can add clauses.

    Examples of such classes include `CNF`, `WCNF`, and the various `Solver`
        implementations.
    """

    def __iadd__(self, constraint: Constraint):
        self.add_encoding(constraint)
        return self

    @abstractmethod
    def add_clause(self, clause: Iterable[Lit]):
        """Add a clause to the database.

        :param clause: An iterable of literals representing the clause to add
        :raises Unsatisfiable: If the formula has become unsatisfiable
        """
        ...

    def add_encoding(
        self,
        constraint: Constraint,
        encoder: Optional[Encoder] = None,
        conditions: Optional[Iterable[Lit]] = None,
    ):
        """Add an encoding of a `constraint` to the database.

        Optionally, the constraint is implied by the given `conditions` (i.e. every
            clause is extended by the `conditions`), and the given `encoder` is used
            for the encoding.

        :param constraint: The constraint or formula to encode and add to the database
        :raises Unsatisfiable: If the formula has become unsatisfiable
        """
        _wrap_encode_constraint(self, constraint, encoder, conditions)

    def new_var(self):
        """Add a new variable to the database."""
        r = self.new_var_range(1)
        assert r.start() == r.end()
        return r.start()

    def new_vars(self, n: int) -> Iterable[Lit]:
        """Add `n` new variables to the database.

        :param n: The number of new variables
        :return: The new variables returned as literals
        """
        return self.new_var_range(n)

    @abstractmethod
    def new_var_range(self, n: int) -> VarRange:
        """Add a continuous range of `n` new variables to the database.

        :param n: The number of new variables
        :return: The start and end of the range of the new variables (inclusive), given
            as literals.
        """
        ...


class CNF(ClauseDatabase):
    """A representation for Boolean formulas in conjunctive normal form."""

    _inner: CNFInner

    def __init__(self):
        self._inner = CNFInner()

    def add_clause(self, clause: Iterable[Lit]):
        return self._inner.add_clause(iter(clause))

    def add_encoding(
        self,
        constraint: Constraint,
        encoder: Optional[Encoder] = None,
        conditions: Optional[Iterable[Lit]] = None,
    ):
        conditions = list(conditions) if conditions is not None else []
        return self._inner.add_encoding(constraint, encoder, conditions)

    def clauses(self) -> Iterable[list[Lit]]:
        """Returns an iterator of the clauses currently included in the CNF.

        :return: An iterable of lists of literals representing the clauses.
        """
        return self._inner.clauses()

    def new_var_range(self, n: int) -> VarRange:
        return self._inner.new_var_range(n)

    def to_dimacs(self) -> str:
        """Return a textual representation in the DIMACS format.

        :return: The CNF as a DIMACS string
        """
        return self._inner.to_dimacs()

    def variables(self) -> Iterable[Lit]:
        """Returns an iterator of the variables currently included in the CNF.

        :return: An iterable of literals representing the variables.
        """
        return self._inner.variables()


class WCNF(CNF):
    """A representation for conjunctive normal form with weighted clauses.

    Note that `WCNF.clauses` only iterates over the hard clauses. Use
    `WCNF.weighted_clauses` to iterate over all clauses.
    """

    _inner: WCNFInner

    def __init__(self):
        self._inner = WCNFInner()

    def add_weighted_clause(self, clause: Iterable[Lit], weight: int):
        """Add a weighted clause to the database.

        :param clause: An iterable of literals representing the clause to add
        :param weight: the weight of the clause
        """
        return self._inner.add_weighted_clause(iter(clause), weight)

    def weighted_clauses(self) -> Iterable[tuple[Optional[int], list[Lit]]]:
        """Returns an iterator of the weighted clauses currently included in the WCNF.

        :return: An iterable of lists of literals representing the clauses.
        """
        return self._inner.weighted_clauses()



import pindakaas


def test_cadical():
    slv = pindakaas.solver.CaDiCaL()
    x, y = slv.new_vars(2)
    slv.add_clause([x, y])
    slv.add_clause([~x, ~y])
    with slv.solve() as result:
        assert result.status == pindakaas.solver.Status.SATISFIED
        vx, vy = result.value(x), result.value(y)
        assert vx is not None
        assert vy is not None
        assert vx != vy


def test_assumptions():
    slv = pindakaas.solver.CaDiCaL()
    x, y = slv.new_vars(2)
    slv += x ^ y
    with slv.solve(assumptions=[x]) as result:
        assert result.status == pindakaas.solver.Status.SATISFIED
        assert result.value(x) is True
        assert result.value(y) is False
    with slv.solve(assumptions=[y]) as result:
        assert result.status == pindakaas.solver.Status.SATISFIED
        assert result.value(x) is False
        assert result.value(y) is True
    with slv.solve(assumptions=[x, y]) as result:
        assert result.status == pindakaas.solver.Status.UNSATISFIABLE
        assert result.failed(x) or result.failed(y), (
            "One or the other variable should have been used to prove UNSAT"
        )


def test_kissat():
    slv = pindakaas.solver.Kissat()
    x, y = slv.new_vars(2)
    slv.add_clause([x, y])
    slv.add_clause([~x, ~y])
    with slv.solve() as result:
        assert result.status == pindakaas.solver.Status.SATISFIED
        vx, vy = result.value(x), result.value(y)
        assert vx is not None
        assert vy is not None
        assert vx != vy


def test_issue_159():
    slv = pindakaas.solver.CaDiCaL()
    assert len(slv.new_vars(1)) == 1
